#include <stdlib.h>
#include <stdio.h>
#include <dos.h>
#include <int.h>

#include "kernel.h"

#define MASTER		0x20
#define SLAVE		0xa0
#define EOI			0x20
#define MASK(pic)	((pic) + 1)
#define	BIT(n)		(1 << ((n) & 7))
#define NSTACKS		6
#define STACKSIZE	2048

typedef struct
{
	void *				stack_bottom;
	unsigned			save_esp;
	unsigned			save_ss;
	unsigned			irq;
}
IntControl_t;

unsigned mt_int_level;

static unsigned char old_mask_master, old_mask_slave;
static IntControl_t int_control[16];
static Handler_t handler[16];
static bool timer_protected;


/*
--------------------------------------------------------------------------------
Conversions between IRQ number and vector number. This is the original PC
setting, change these functions if the PICS are reprogrammed.
--------------------------------------------------------------------------------
*/

static unsigned
vec_num(unsigned irq)
{
	if ( irq < 8 )
		return irq + 8;
	return irq - 8 + 0x70;
}

static unsigned
irq_num(unsigned vec)
{
	if ( vec < 0x70 )
		return vec - 8;
	return vec - 0x70 + 8;
}


/*
--------------------------------------------------------------------------------
Generic HW interrupt vector. This is installed for each interrupt hooked using
int_intercept(). As interrupts may reenter, stack handling is done here. The
first NSTACKS nested interrupts will get an internal stack of STACKSIZE bytes.
When returning from a first level interrupt, if it happened in protected mode
and on the application code (CS == our CS), the return stack and address
are changed so that the interrupt will return to Scheduler(), which will itself
return to the point of origin. This is the way to implement preemptions; it only
works with X32 running on plain DOS, in Windows it is always CS != our CS. 
This vector returns true if the interrupt was handled completely, or false in
order to chain to the previous handler.
--------------------------------------------------------------------------------
*/

__declspec(naked)
static bool
vector(struct INT_DATA *d)
{
	static IntControl_t *ctl;
	static unsigned vec;
	static bool complete;
	
	asm
	{
		push	ebp
		mov		ebp, esp
		mov		eax, [ebp+40]
		mov		cx, ss:[eax+12]						// Vector number
		mov		word ptr vec, cx
	}

	ctl = &int_control[mt_int_level];
	ctl->irq = irq_num(vec);
	
	if ( ctl->stack_bottom )						// Change to local stack
		asm
		{
			mov		eax, ctl
			mov		[eax]IntControl_t.save_esp, esp
			mov		word ptr [eax]IntControl_t.save_ss, ss
			mov		esp, [eax]IntControl_t.stack_bottom
			mov		ss, word ptr mt_ds
		}

	mt_int_level++;
	handler[ctl->irq](ctl->irq);					// Call installed handler
	disable();										// (might enable ints.)
	asm	mov complete, eax;							// Completion status
	mt_int_level--;
		
	ctl = &int_control[mt_int_level];

	if ( ctl->stack_bottom )						// Back to original stack
		asm
		{
			mov		eax, ctl
			mov		esp, [eax]IntControl_t.save_esp
			mov		ss, word ptr [eax]IntControl_t.save_ss
		}

	if ( !mt_int_level )							// 1st level return
		asm
		{
			mov		eax, [ebp+40]
			mov		cx, ss:[eax+14]					// Check
			cmp		cx, 0							// mode == protected
			jne		returning
			mov		cx, ss:[eax+20]					// Check
			cmp		cx, word ptr mt_cs				// return CS == CS		
			jne		returning
			mov		ecx, ss:[eax+16]				// Original return EIP
			mov		ss:[eax+16], offset Scheduler	// replaced by Scheduler
			sub		ss:[eax+28], 4					// Return ESP -= 4
			les		edx, ss:[eax+28]				// Scheduler will return
			mov		es:[edx], ecx					// to original EIP
		}

	returning:
	asm
	{
		mov		eax, complete						// If false, will chain to
		mov		esp, ebp							// previous handler
		pop		ebp
		ret
	}
}

/*
--------------------------------------------------------------------------------
Module initialization and cleanup
--------------------------------------------------------------------------------
*/

void
mt_restore_irqs(void)
{
	int i;

	DisableInts();
	_outp(MASK(MASTER), old_mask_master);
	_outp(MASK(SLAVE), old_mask_slave);
	timer_protected = false;
	for ( i = 0 ; i < 16 ; i++ )
		SetHandler(i, NULL);
	RestoreInts();
}

void
mt_setup_irqs(void)
{	
	int i;

	old_mask_master = _inp(MASK(MASTER));
	old_mask_slave = _inp(MASK(SLAVE));
	for ( i = 0 ; i < NSTACKS ; i++ )
		int_control[i].stack_bottom = (char *) Malloc(STACKSIZE) + STACKSIZE;
}

/*
--------------------------------------------------------------------------------
SetHandler - set/restore a handler for a hardware interrupt

Passing NULL for the handler restores the original vector.
For the timer interrupt this can only be called once, which is done by the
kernel, and no further trapping or untrapping is possible.
--------------------------------------------------------------------------------
*/

void
SetHandler(unsigned irq, Handler_t isr)
{
	if ( irq == CLOCKIRQ )
	{
		if ( timer_protected )
			Panic("SetHandler: timer interrupt not allowed");
		timer_protected = true;
	}

	DisableInts();
	if ( !handler[irq] && isr )
		int_intercept(vec_num(irq), vector, 0);
	else if ( handler[irq] && !isr )
		int_restore(vec_num(irq));
	handler[irq] = isr;
	RestoreInts();
}

/*
--------------------------------------------------------------------------------
DisableIRQ, EnableIRQ - enable/disable individual IRQs
--------------------------------------------------------------------------------
*/

void
DisableIRQ(unsigned irq)
{
	DisableInts();
	if ( irq <= 7 )
		_outp(MASK(MASTER), _inp(MASK(MASTER)) | BIT(irq));
	else
		_outp(MASK(SLAVE), _inp(MASK(SLAVE)) | BIT(irq));
	RestoreInts();
}

void
EnableIRQ(unsigned irq)
{
	DisableInts();
	if ( irq <= 7 )
		_outp(MASK(MASTER), _inp(MASK(MASTER)) & ~BIT(irq));
	else
		_outp(MASK(SLAVE), _inp(MASK(SLAVE)) & ~BIT(irq));
	RestoreInts();
}

/*
--------------------------------------------------------------------------------
EndInterrupt - send EOI to the PICs

Must be called from interrupt handler to reenable interrupt, unless control 
is chained to the previous handler which takes care of that.
--------------------------------------------------------------------------------
*/

void
EndInterrupt(void)
{
	IntControl_t *ctl = &int_control[mt_int_level-1];

	if ( ctl->irq > 7 )								// EOI to slave PIC
		_outp(SLAVE, EOI);
	_outp(MASTER, EOI);								// EOI to master PIC
}

