#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "mtask.h"
#include "msgqueue.h"


/*
--------------------------------------------------------------------------------
Get_msg, put_msg - internal functions to get/put a message

These are standard producer/consumer circular buffer functions which are not 
serialized, i.e. they work properly for only one producer and only one consumer.
--------------------------------------------------------------------------------
*/

static bool
get_msg(MsgQueue_t *mq, void *msg, Time_t msecs)
{
	if ( !WaitSemTimed(mq->sem_get, msecs) )
		return false;
	memcpy(msg, mq->head, mq->msg_size);
	mq->head += mq->msg_size;
	if ( mq->head == mq->end )
		mq->head = mq->buf;
	SignalSem(mq->sem_put);
	return true;
}

static bool
put_msg(MsgQueue_t *mq, void *msg, Time_t msecs)
{
	if ( !WaitSemTimed(mq->sem_put, msecs) )
		return false;
	memcpy(mq->tail, msg, mq->msg_size);
	mq->tail += mq->msg_size;
	if ( mq->tail == mq->end )
		mq->tail = mq->buf;
	SignalSem(mq->sem_get);
	return true;
}

/*
--------------------------------------------------------------------------------
CreateMsgQueue, DeleteMsgQueue - create/delete a message queue

The msg_max parameter is que queue capacity (max. quantity of messages to store)
and msg_size the message size. The two booleans indicate whether input/output
should be serialized in case there are several message producers/consumers.
--------------------------------------------------------------------------------
*/

MsgQueue_t *
CreateMsgQueue(char *name, unsigned msg_max, unsigned msg_size,
					bool serialized_get, bool serialized_put)
{
	char buf[200];
	MsgQueue_t *mq;
	unsigned size = msg_max * msg_size;

	if ( size / msg_size != msg_max )	
		Panic("CreateMsgQueue: size overflow");

	mq = Malloc(sizeof(MsgQueue_t));
	mq->msg_size = msg_size;
	mq->head = mq->tail = mq->buf = Malloc(size);
	mq->end = mq->buf + size;
	sprintf(buf, "Get %s", name);
	mq->sem_get = CreateSem(buf, 0);
	if ( serialized_get )
		mq->mutex_get = CreateMutex(buf);
	sprintf(buf, "Put %s", name);
	mq->sem_put = CreateSem(buf, msg_max);
	if ( serialized_put )
		mq->mutex_put = CreateMutex(buf);

	return mq;
}

void
DeleteMsgQueue(MsgQueue_t *mq)
{
	DeleteSem(mq->sem_get);
	if ( mq->mutex_get )
		DeleteMutex(mq->mutex_get);
	DeleteSem(mq->sem_put);
	if ( mq->mutex_put )
		DeleteMutex(mq->mutex_put);
	Free(mq->buf);
	Free(mq);
}

/*
--------------------------------------------------------------------------------
GetMsgQueue, GetMsgQueueCond, GetMsgQueueTimed - get a message from a queue
--------------------------------------------------------------------------------
*/

bool
GetMsgQueue(MsgQueue_t *mq, void *msg)
{
	return GetMsgQueueTimed(mq, msg, FOREVER);
}

bool
GetMsgQueueCond(MsgQueue_t *mq, void *msg)
{
	return GetMsgQueueTimed(mq, msg, 0);
}

bool
GetMsgQueueTimed(MsgQueue_t *mq, void *msg, Time_t msecs)
{
	bool result;

	if ( mq->mutex_get && !EnterMutexTimed(mq->mutex_get, msecs) )
		return false;
	result = get_msg(mq, msg, msecs);
	if ( mq->mutex_get )
		LeaveMutex(mq->mutex_get);

	return result;
}

/*
--------------------------------------------------------------------------------
PutMsgQueue, PutMsgQueueCond, PutMsgQueueTimed - put a message in the queue
--------------------------------------------------------------------------------
*/

bool
PutMsgQueue(MsgQueue_t *mq, void *msg)
{
	return PutMsgQueueTimed(mq, msg, FOREVER);
}

bool
PutMsgQueueCond(MsgQueue_t *mq, void *msg)
{
	return PutMsgQueueTimed(mq, msg, 0);
}

bool
PutMsgQueueTimed(MsgQueue_t *mq, void *msg, Time_t msecs)
{
	bool result;

	if ( mq->mutex_put && !EnterMutexTimed(mq->mutex_put, msecs) )
		return false;
	result = put_msg(mq, msg, msecs);
	if ( mq->mutex_put )
		LeaveMutex(mq->mutex_put);

	return result;
}

/*
--------------------------------------------------------------------------------
AvailMsgQueue - tells how many messages are stored in the queue
--------------------------------------------------------------------------------
*/

unsigned
AvailMsgQueue(MsgQueue_t *mq)
{
	return ValueSem(mq->sem_get);
}
