#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <disp.h>
#include <string.h>
#include <disp.h>
#include <math.h>
#include "mtask.h"
#include "mutex.h"
#include "monitor.h"

#define NF						5
#define MIN_THINK				2000
#define MAX_THINK				6000
#define MIN_EAT					2000
#define MAX_EAT					6000

enum COLORS
{
    BLACK,          /* dark colors */
    BLUE,
    GREEN,
    CYAN,
    RED,
    MAGENTA,
    BROWN,
    LIGHTGRAY,
    DARKGRAY,       /* light colors */
    LIGHTBLUE,
    LIGHTGREEN,
    LIGHTCYAN,
    LIGHTRED,
    LIGHTMAGENTA,
    YELLOW,
    WHITE
};

#define COLOR_NORMAL			LIGHTGRAY
#define COLOR_THINKING			WHITE
#define COLOR_HUNGRY			YELLOW
#define COLOR_EATING			LIGHTRED
#define COLOR_FREE				WHITE
#define COLOR_USED				LIGHTRED

typedef struct
{
	int x;
	int y;
}
point;

Mutex_t *mut;					// Serialize DOS and rand()
Monitor_t *mon;					// Monitor for accessing forks
bool in_use[NF];				// Used forks
Condition_t *cond[NF];			// One condition variable per philosopher

int
prev(int n)
{
	return n == 0 ? NF-1 : n-1;
}

int
next(int n)
{
	return n == NF-1 ? 0 : n+1;
}

#define left_fork(i)			(i)
#define right_fork(i)			next(i)
#define left_phil(i)			prev(i)
#define right_phil(i)			next(i)

point pos[2*NF];				// Screen coordinates of philosophers and forks

#define phil_position(i)		(pos[2*(i)+1])
#define fork_position(i)		(pos[2*(i)])

int 
mprint(point where, char *msg, unsigned color)
{
	int n;
	static int width;

	EnterMutex(mut);
	if ( (n = strlen(msg)) > width )
		width = n;
	disp_move(where.y-1, where.x-1);
	disp_setattr(color);
	n = disp_printf("%-*.*s", width, width, msg);
	LeaveMutex(mut);

	return n;
}

void 
random_delay(int d1, int d2)
{
	Time_t t;

	EnterMutex(mut);
	t = d1 + rand() % (d2-d1);
	LeaveMutex(mut);

	Delay(t);
}

void
think(int i)
{
	mprint(phil_position(i), "THINKING", COLOR_THINKING);
	random_delay(MIN_THINK, MAX_THINK);
}

void
become_hungry(int i)
{
	mprint(phil_position(i), "HUNGRY", COLOR_HUNGRY);
}

void
eat(int i)
{
	mprint(phil_position(i), "EATING", COLOR_EATING);
	random_delay(MIN_EAT, MAX_EAT);
}

bool
forks_avail(int i)
{
	return !in_use[left_fork(i)] && !in_use[right_fork(i)];
}

void
take_forks(int i)
{
	in_use[left_fork(i)] = in_use[right_fork(i)] = true;
	mprint(fork_position(left_fork(i)), "used", COLOR_USED);
	mprint(fork_position(right_fork(i)), "used", COLOR_USED);
}

void
leave_forks(int i)
{
	in_use[left_fork(i)] = in_use[right_fork(i)] = false;
	mprint(fork_position(left_fork(i)), "free", COLOR_FREE);
	mprint(fork_position(right_fork(i)), "free", COLOR_FREE);
}

void
philosopher(void *arg)
{
	int i;
	unsigned len = sizeof i;
	int left, right;

	Receive(NULL, &i, &len);
	left = left_phil(i);
	right = right_phil(i);

	while ( true )
	{
		think(i);
		become_hungry(i);
		EnterMonitor(mon);
		while ( !forks_avail(i) )
			WaitCondition(cond[i]);
		take_forks(i);
		LeaveMonitor(mon);
		eat(i);
		EnterMonitor(mon);
		leave_forks(i);
		if ( forks_avail(left) )
			SignalCondition(cond[left]);
		if ( forks_avail(right) )
			SignalCondition(cond[right]);
		LeaveMonitor(mon);
	}
}

void 
set_positions(void)
{
	int i;

	for ( i = 0 ; i < 2*NF ; i++ )
	{
		pos[i].x = 35 + 20 * cos(i * M_PI/NF);
		pos[i].y = 14 - 10 * sin(i * M_PI/NF);
	}
}

void
clrscr(void)
{
	disp_move(0, 0);
	disp_eeop();
	disp_move(0, 0);
}

void 
restore(void)
{
	disp_setattr(COLOR_NORMAL);
	disp_showcursor();
	clrscr();
}

int 
main(void)
{
	int i;
	Task_t *t;

	InitMTask(1, 100);

	disp_hidecursor();
	clrscr();
	
	atexit(restore);

	disp_puts("Press any key to quit...");

	// Initialize data structures
	mut = CreateMutex(NULL);
	mon = CreateMonitor(NULL);
	set_positions();
	for ( i = 0 ; i < NF ; i++ )
	{
		mprint(fork_position(i), "free", COLOR_FREE);
		cond[i] = CreateCondition(NULL, mon);
	}

	// Launch tasks
	srand(time(NULL));
	for ( i = 0 ; i < NF ; i++ )
	{
		t = CreateTask(philosopher, 4000, NULL, NULL, DEFAULT_PRIO);
		Ready(t);
		Send(t, &i, sizeof i);
	}

	while ( true )
	{
		EnterMutex(mut);
		if ( kbhit() )
		{
			getch();
			break;
		}
		LeaveMutex(mut);
	}
	return 0;
}
